#
# This file is part of the pylibssh library
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see file LICENSE.rst in this
# repository.
#
cdef class LibsshException(Exception):
    def __init__(self, message=''):
        self.message = message
        super(LibsshException, self).__init__(message)

    def __str__(self):
        return self.message

    def __repr__(self):
        return self.message

    def _get_session_error_str(self, obj):
        return libssh.ssh_get_error(<void*>obj._libssh_session).decode()


cdef class LibsshSessionException(LibsshException):
    pass


cdef class LibsshChannelException(LibsshException):
    pass


class LibsshChannelReadFailure(LibsshChannelException, ConnectionError):
    """Raised when there is a failure to read from a libssh channel."""


cdef class LibsshSCPException(LibsshException):
    pass


cdef class LibsshSFTPException(LibsshException):
    pass



#
# This file is part of the pylibssh library
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see file LICENSE.rst in this
# repository.
#
import signal
import time
from io import BytesIO

from cpython.bytes cimport PyBytes_AS_STRING
from libc.string cimport memset

from pylibsshext.errors cimport LibsshChannelException
from pylibsshext.errors import LibsshChannelReadFailure
from pylibsshext.session cimport get_libssh_session

from subprocess import CompletedProcess


cdef int _process_outputs(libssh.ssh_session session,
                          libssh.ssh_channel channel,
                          void *data,
                          libssh.uint32_t len,
                          int is_stderr,
                          void *userdata) with gil:
    if len == 0:
        return 0
    data_b = <bytes>(<char *>data)[:len]
    result = <object>userdata
    if is_stderr:
        result.stderr += data_b
    else:
        result.stdout += data_b
    return len

cdef class ChannelCallback:
    def __cinit__(self):
        memset(&self.callback, 0, sizeof(self.callback))
        callbacks.ssh_callbacks_init(&self.callback)
        self.callback.channel_data_function = <callbacks.ssh_channel_data_callback>&_process_outputs

    def set_user_data(self, userdata):
        self._userdata = userdata
        self.callback.userdata = <void *>self._userdata

cdef class Channel:
    def __cinit__(self, session):
        self._session = session
        self._libssh_session = get_libssh_session(session)
        self._libssh_channel = libssh.ssh_channel_new(self._libssh_session)

        if self._libssh_channel is NULL:
            raise MemoryError
        rc = libssh.ssh_channel_open_session(self._libssh_channel)

        if rc != libssh.SSH_OK:
            libssh.ssh_channel_free(self._libssh_channel)
            self._libssh_channel = NULL
            raise LibsshChannelException("Failed to open_session: [%d]" % rc)

    def __dealloc__(self):
        if self._libssh_channel is not NULL:
            libssh.ssh_channel_close(self._libssh_channel)
            libssh.ssh_channel_free(self._libssh_channel)
            self._libssh_channel = NULL

    def request_shell(self):
        self.request_pty()
        rc = libssh.ssh_channel_request_shell(self._libssh_channel)
        if rc != libssh.SSH_OK:
            raise LibsshChannelException("Failed to request_shell: [%d]" % rc)

    def request_exec(self, command):
        """Run a shell command without an interactive shell."""
        rc = libssh.ssh_channel_request_exec(self._libssh_channel, command.encode("utf-8"))
        if rc != libssh.SSH_OK:
            raise LibsshChannelException("Failed to request_exec: [%d]" % rc)

    def request_pty(self):
        rc = libssh.ssh_channel_request_pty(self._libssh_channel)
        if rc != libssh.SSH_OK:
            raise LibsshChannelException("Failed to request pty: [%d]" % rc)

    def request_pty_size(self, terminal, col, row):
        rc = libssh.ssh_channel_request_pty_size(self._libssh_channel, terminal, col, row)
        if rc != libssh.SSH_OK:
            raise LibsshChannelException("Failed to request pty with [%d] for terminal [%s], "
                                         "columns [%d] and rows [%d]" % (rc, terminal, col, row))
        rc = libssh.ssh_channel_request_shell(self._libssh_channel)
        if rc != libssh.SSH_OK:
            raise LibsshChannelException("Failed to request_shell: [%d]" % rc)

    def poll(self, timeout=-1, stderr=0):
        if timeout < 0:
            rc = libssh.ssh_channel_poll(self._libssh_channel, stderr)
        else:
            rc = libssh.ssh_channel_poll_timeout(self._libssh_channel, timeout, stderr)
        if rc == libssh.SSH_ERROR:
            raise LibsshChannelException("Failed to poll channel: [{0}]".format(rc))
        return rc

    def read_nonblocking(self, size=1024, stderr=0):
        cdef char buffer[1024]
        size_m = size
        if size_m > sizeof(buffer):
            size_m = sizeof(buffer)
        nbytes = libssh.ssh_channel_read_nonblocking(self._libssh_channel, buffer, size_m, stderr)
        if nbytes == libssh.SSH_ERROR:
            # This is what Session._get_session_error_str() does, but we don't have the Python object
            error = libssh.ssh_get_error(<void*>self._libssh_session).decode()
            raise LibsshChannelReadFailure(error)
        elif nbytes == libssh.SSH_EOF:
            return None

        return <bytes>buffer[:nbytes]

    def recv(self, size=1024, stderr=0):
        return self.read_nonblocking(size=size, stderr=stderr)

    def write(self, data):
        written = libssh.ssh_channel_write(self._libssh_channel, PyBytes_AS_STRING(data), len(data))
        if written == libssh.SSH_ERROR:
            raise LibsshChannelException("Failed to write to ssh channel")
        return written

    def sendall(self, data):
        return self.write(data)

    def read_bulk_response(self, stderr=0, timeout=0.001, retry=5):
        if retry <= 0:
            raise ValueError(
                'Got arg `retry={arg!r}` but it must be greater than 0'.
                format(arg=retry),
            )

        response = b""
        with BytesIO() as recv_buff:
            for _ in range(retry, 0, -1):
                data = self.read_nonblocking(size=1024, stderr=stderr)
                if not data:
                    if timeout:
                        time.sleep(timeout)
                    continue

                recv_buff.write(data)
            response = recv_buff.getvalue()
        return response

    def exec_command(self, command):
        # request_exec requires a fresh channel each run, so do not use the existing channel
        cdef libssh.ssh_channel channel = libssh.ssh_channel_new(self._libssh_session)
        if channel is NULL:
            raise MemoryError

        rc = libssh.ssh_channel_open_session(channel)
        if rc != libssh.SSH_OK:
            libssh.ssh_channel_free(channel)
            raise LibsshChannelException("Failed to open_session: [{0}]".format(rc))

        result = CompletedProcess(args=command, returncode=-1, stdout=b'', stderr=b'')

        cb = ChannelCallback()
        cb.set_user_data(result)
        callbacks.ssh_set_channel_callbacks(channel, &cb.callback)
        # keep the callback around in the session object to avoid use after free
        self._session.push_callback(cb)

        rc = libssh.ssh_channel_request_exec(channel, command.encode("utf-8"))
        if rc != libssh.SSH_OK:
            libssh.ssh_channel_close(channel)
            libssh.ssh_channel_free(channel)
            raise LibsshChannelException("Failed to execute command [{0}]: [{1}]".format(command, rc))

        # wait before remote writes all data before closing the channel
        while not libssh.ssh_channel_is_eof(channel):
            libssh.ssh_channel_poll(channel, 0)

        libssh.ssh_channel_send_eof(channel)
        result.returncode = libssh.ssh_channel_get_exit_status(channel)
        if channel is not NULL:
            libssh.ssh_channel_close(channel)
            libssh.ssh_channel_free(channel)

        return result

    def send_eof(self):
        """Send EOF to the channel, this will close stdin."""
        rc = libssh.ssh_channel_send_eof(self._libssh_channel)
        if rc != libssh.SSH_OK:
            raise LibsshChannelException("Failed to ssh_channel_send_eof: [%d]" % rc)

    def send_signal(self, sig):
        """
        Send signal to the remote process.

        :param sig: a signal constant from ``signal``, e.g. ``signal.SIGUSR1``.
        :type sig: signal.Signals
        """
        if not isinstance(sig, signal.Signals):
            raise TypeError(f"Expecting signal.Signals not {type(sig)}")

        sshsig = sig.name.replace("SIG", "")  # FIXME: replace w/ `str.removeprefix()` once Python 3.8 support is dropped
        rc = libssh.ssh_channel_request_send_signal(self._libssh_channel, sshsig.encode("utf-8"))
        if rc != libssh.SSH_OK:
            raise LibsshChannelException("Failed to ssh_channel_request_send_signal: [%d]" % rc)

    def get_channel_exit_status(self):
        return libssh.ssh_channel_get_exit_status(self._libssh_channel)

    @property
    def is_eof(self):
        """True if remote has sent an EOF."""
        rc = libssh.ssh_channel_is_eof(self._libssh_channel)
        return rc != 0

    def close(self):
        if self._libssh_channel is not NULL:
            libssh.ssh_channel_close(self._libssh_channel)
            libssh.ssh_channel_free(self._libssh_channel)
            self._libssh_channel = NULL



#
# This file is part of the pylibssh library
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see file LICENSE.rst in this
# repository.

import os

from cpython.bytes cimport PyBytes_AS_STRING
from cpython.mem cimport PyMem_Free, PyMem_Malloc

from pylibsshext.errors cimport LibsshSCPException
from pylibsshext.session cimport get_libssh_session


SCP_MAX_CHUNK = 65_536  # 64kB


cdef class SCP:
    def __cinit__(self, session):
        self.session = session
        self._libssh_session = get_libssh_session(session)

    def put(self, local_file, remote_file):
        """
        Create an SCP channel and send a file to the remote host over that channel.

        :param local_file: The path on the local host where the file will be read from
        :type local_file: str

        :param remote_file: The path on the remote host where the file should be placed
        :type remote_file: str
        """
        remote_file_b = remote_file
        if isinstance(remote_file_b, unicode):
            remote_file_b = remote_file.encode("utf-8")
        remote_dir_b, filename_b = os.path.split(remote_file_b)
        if not remote_dir_b:
            remote_dir_b = b"."

        with open(local_file, "rb") as f:
            file_stat = os.fstat(f.fileno())
            file_size = file_stat.st_size
            file_mode = file_stat.st_mode & 0o777

            # Create the SCP session in write mode
            scp = libssh.ssh_scp_new(self._libssh_session, libssh.SSH_SCP_WRITE, remote_file_b)
            if scp is NULL:
                raise LibsshSCPException(
                    "Allocating SCP session of remote file [{path!s}] for "
                    "write failed with error [{err!s}]".
                    format(path=remote_file, err=self._get_ssh_error_str()),
                )

            # Initialize the SCP channel
            rc = libssh.ssh_scp_init(scp)
            if rc != libssh.SSH_OK:
                libssh.ssh_scp_free(scp)
                raise LibsshSCPException(
                    "Initializing SCP session of remote file [{path!s}] for "
                    "write failed with error [{err!s}]".
                    format(path=remote_file, err=self._get_ssh_error_str()),
                )

            try:
                # Read buffer
                read_buffer_size = min(file_size, SCP_MAX_CHUNK)

                # Begin to send to the file
                rc = libssh.ssh_scp_push_file(scp, filename_b, file_size, file_mode)
                if rc != libssh.SSH_OK:
                    raise LibsshSCPException("Can't open remote file: %s" % self._get_ssh_error_str())

                remaining_bytes_to_read = file_size
                while remaining_bytes_to_read > 0:
                    # Read the chunk from local file
                    read_bytes = min(remaining_bytes_to_read, read_buffer_size)
                    read_buffer = f.read(read_bytes)

                    # Write to the open file
                    rc = libssh.ssh_scp_write(scp, PyBytes_AS_STRING(read_buffer), read_bytes)
                    if rc != libssh.SSH_OK:
                        raise LibsshSCPException("Can't write to remote file: %s" % self._get_ssh_error_str())
                    remaining_bytes_to_read -= read_bytes
            finally:
                libssh.ssh_scp_close(scp)
                libssh.ssh_scp_free(scp)

            return libssh.SSH_OK

    def get(self, remote_file, local_file):
        """
        Create an SCP channel and retrieve a file from the remote host over that channel.

        :param remote_file: The path on the remote host where the file will be read from
        :type remote_file: str

        :param local_file: The path on the local host where the file should be placed
        :type local_file: str
        """
        cdef char *read_buffer = NULL

        remote_file_b = remote_file
        if isinstance(remote_file_b, unicode):
            remote_file_b = remote_file.encode("utf-8")

        # Create the SCP session in read mode
        scp = libssh.ssh_scp_new(self._libssh_session, libssh.SSH_SCP_READ, remote_file_b)
        if scp is NULL:
            raise LibsshSCPException("Allocating SCP session of remote file [%s] for write failed with error [%s]" % (remote_file, self._get_ssh_error_str()))

        # Initialize the SCP channel
        rc = libssh.ssh_scp_init(scp)
        if rc != libssh.SSH_OK:
            libssh.ssh_scp_free(scp)
            raise LibsshSCPException("Initializing SCP session of remote file [%s] for write failed with error [%s]" % (remote_file, self._get_ssh_error_str()))

        try:
            # Request to pull the file
            rc = libssh.ssh_scp_pull_request(scp)
            if rc != libssh.SSH_SCP_REQUEST_NEWFILE:
                raise LibsshSCPException("Error receiving information about file: %s" % self._get_ssh_error_str())

            size = libssh.ssh_scp_request_get_size(scp)
            mode = libssh.ssh_scp_request_get_permissions(scp)

            # cap the buffer size to reasonable number -- libssh will not return the whole data at once anyway
            read_buffer_size = min(size, SCP_MAX_CHUNK)
            read_buffer = <char *>PyMem_Malloc(read_buffer_size)
            if read_buffer is NULL:
                raise LibsshSCPException("Memory allocation error")

            # Indicate the transfer is ready to begin
            libssh.ssh_scp_accept_request(scp)
            if rc == libssh.SSH_ERROR:
                raise LibsshSCPException("Failed to start read request: %s" % self._get_ssh_error_str())

            remaining_bytes_to_read = size
            with open(local_file, "wb") as f:
                while remaining_bytes_to_read > 0:
                    requested_read_bytes = min(remaining_bytes_to_read, read_buffer_size)
                    read_bytes = libssh.ssh_scp_read(scp, read_buffer, requested_read_bytes)
                    if read_bytes == libssh.SSH_ERROR:
                        raise LibsshSCPException("Error receiving file data: %s" % self._get_ssh_error_str())

                    py_file_bytes = read_buffer[:read_bytes]
                    f.write(py_file_bytes)
                    remaining_bytes_to_read -= read_bytes
            if mode >= 0:
                os.chmod(local_file, mode)

            # Make sure we have finished requesting files
            rc = libssh.ssh_scp_pull_request(scp)
            if rc != libssh.SSH_SCP_REQUEST_EOF:
                raise LibsshSCPException("Unexpected request: %s" % self._get_ssh_error_str())

        finally:
            if read_buffer is not NULL:
                PyMem_Free(read_buffer)
            libssh.ssh_scp_close(scp)
            libssh.ssh_scp_free(scp)

        return libssh.SSH_OK

    def _get_ssh_error_str(self):
        return libssh.ssh_get_error(self._libssh_session)
